//===----------------------------------------------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

// INVOKE (f, t1, t2, ..., tN)

// The tests for INVOKE (f, t1, t2, ..., tN) live in the "test/libcxx" tree
// since they require calling the implementation specific "__invoke" and
// "__invoke" functions.

int main(int, char**)
{
  return 0;
}
