#define FLT_RADIX 2

#define FLT_MIN        1.1754943e-38f
#define FLT_MAX        3.4028234e38f
#define FLT_MIN_EXP    -125
#define FLT_MIN_10_EXP -37
#define FLT_MAX_EXP    128
#define FLT_MAX_10_EXP 38
#define FLT_MANT_DIG   24
#define FLT_EPSILON    __FLT_EPSILON__
#define FLT_DIG        6

#define DBL_MIN        2.2250738585072013e-308
#define DBL_MAX        1.7976931348623157e308
#define DBL_MIN_EXP    -1021
#define DBL_MIN_10_EXP -307
#define DBL_MAX_EXP    1024
#define DBL_MAX_10_EXP 308
#define DBL_MANT_DIG   53
#define DBL_EPSILON    __DBL_EPSILON__
#define DBL_DIG        15
